const axios = require('axios');
const ethers = require('ethers');
const crypto = require('crypto');
const objectHash = require('object-hash');

const { getContractAddressByNetworkName } = require('./contract-config');
const { authorAddresses, authorPrivKeys, forum } = require('./db');

const getContract = (name) => ethers.getContractAt(
  name,
  getContractAddressByNetworkName(process.env.NETWORK, name),
);

const fetchPaperInfo = async (paperId) => {
  const paper = await axios.get(`https://api.semanticscholar.org/graph/v1/paper/${paperId}`, {
    headers: {
      'api-key': process.env.SEMANTIC_SCHOLAR_API_KEY,
    },
  });
  return paper;
};

const getAuthorsInfo = async (paper) => Promise.all(paper.authors.map(async ({ authorId }) => {
  // Check if we already have an account for each author
  let authorAddress;
  let authorPrivKey;
  try {
    authorAddress = await authorAddresses.get(authorId);
  } catch (e) {
    // Probably not found
  }
  if (authorAddress) {
    // This should always succeed, so we don't use try/catch here
    authorPrivKey = await authorPrivKeys.get(authorAddress);
  } else {
    // Generate and store a new account
    const id = crypto.randomBytes(32).toString('hex');
    authorPrivKey = `0x${id}`;
    const wallet = new ethers.Wallet(authorPrivKey);
    authorAddress = wallet.address;
    await authorAddress.put(authorId, authorAddress);
    await authorPrivKeys.put(authorAddress, authorPrivKey);
  }
  return {
    authorAddress,
    authorPrivKey,
  };
}));

const generatePost = async (paper) => {
  const authorsInfo = getAuthorsInfo(paper);
  const firstAuthorWallet = new ethers.Wallet(authorsInfo[0].authorPrivKey);
  const eachAuthorWeightPercent = Math.floor(100 / authorsInfo.length);
  const authors = authorsInfo.map(({ authorAddress }) => ({
    weightPercent: eachAuthorWeightPercent,
    authorAddress,
  }));
    // Make sure author weights sum to 100
  const totalAuthorsWeight = authors.reduce((t, { weightPercent }) => t + weightPercent);
  authors[0].weightPercent += 100 - totalAuthorsWeight;

  const content = `Semantic Scholar paper ${paper.paperId}
${paper.title}
HREF ${paper.url}`;

  // Note that for now we leave embedded data empty, but the stub is here in case we want to use it
  const embeddedData = {};
  let contentToSign = content;
  if (embeddedData && Object.entries(embeddedData).length) {
    contentToSign += `\n\nDATA\n${JSON.stringify(embeddedData, null, 2)}`;
  }
  const signature = firstAuthorWallet.signMessageSync(contentToSign);
  const hash = objectHash({
    authors, content, signature, embeddedData,
  });
  return {
    hash, authors, content, signature, embeddedData,
  };
};

module.exports = async (req, res) => {
  const dao = await getContract('DAO');
  const {
    body: {
      paperId,
    },
  } = req;

  console.log(`importFromSS ${paperId}`);

  // Read the paper info from SS
  const paper = await fetchPaperInfo(paperId);

  const citations = [];

  if (paper.references) {
    const eachCitationWeightPercent = Math.floor(30 / paper.references.length);
    paper.references.forEach(async ({ paperId: citedPaperId }) => {
      // We need to fetch this paper so we can
      // We need to generate the post we would add to the forum, sign, and hash it.
      // The forum allows dangling citations to support this use case.
      const citedPaper = await fetchPaperInfo(citedPaperId);
      const citedPaperInfo = await generatePost(citedPaper);

      citations.push({
        weightPercent: eachCitationWeightPercent,
        targetPostId: citedPaperInfo.hash,
      });
    });
    // Make sure citation weights sum to 100
    const totalCitationWeight = citations.reduce((t, { weightPercent }) => t + weightPercent);
    citations[0].weightPercent += 100 - totalCitationWeight;
  }

  // Create a post for this paper
  const {
    hash, authors, content, signature, embeddedData,
  } = await generatePost(paper);

  // Write the new post to our database
  await forum.put(hash, {
    authors, content, signature, embeddedData, citations,
  });

  // Add the post to the form (on-chain)
  await dao.addPost(authors, hash, citations);

  console.log({
    authors, content, signature, embeddedData, citations,
  });

  res.end();
};
